/**
 *  Copyright (C) 2013
 *
 *  	 Implementation
 *
 * This is the implementation of the first version of remote invalidation.
 * It's an extension of TLI by executing commit at remote server.
 *
 */

#include "algs.hpp"
#include "RedoRAWUtils.hpp"
#include <sys/mman.h>

// Don't just import everything from stm. This helps us find bugs.
using stm::TxThread;
using stm::timestamp;
using stm::client_cpuset;
using stm::main_server_cpuset;
using stm::server_initialized;
using stm::num_cores;
using stm::clients_exist;
using stm::threadcount;
using stm::threads;
using stm::WriteSetEntry;
using stm::ValueList;
using stm::ValueListEntry;
using stm::filter_t;
using stm::pad_word_t;

/*
 *  RIV1 definitions
 */

#define PAGE_SIZE	4096
#define CACHE_LINE_SIZE 64
#define r_align(n, r)        (((n) + (r) - 1) & -(r))
#define cache_align(n)       r_align(n , CACHE_LINE_SIZE)
#define pad_to_cache_line(n) (cache_align(n) - (n))

// Server-Client Request states
#define READY 		0
#define PENDING		1
#define ABORTED		2

#define GET_NODES_CMD																										\
		"NODES=/sys/devices/system/node;"																			\
		"CPUS=/sys/devices/system/cpu;"																				\
		"if [ -d $NODES ]; then"																							\
	  "  for F in `ls -d $NODES/node*`; do if [ $(cat $F/cpulist | grep '-') ]; then seq -s ' ' $(cat $F/cpulist | tr -s '-' ' '); else cat $F/cpulist | tr -s ',' ' '; fi; done;" \
		"else"																																\
		"  find $CPUS -maxdepth 1 -name \"cpu[0-9]*\" -exec basename {} \\; | sed -e 's/cpu\\(.*\\)/\\1/g' | tr -s '\n' ' ';" \
		"fi"

/*
 *  RIV1 structures
 */
struct request {
	volatile int state; /* the state of commit request */
	TxThread* tx; /* Tx context */
	char pad[pad_to_cache_line(sizeof(int) + sizeof(TxThread*))];
};

/*
 * RIV1 Variables
 */

static struct request* requests = 0; /* requests array */

namespace {

const uintptr_t VALIDATION_FAILED = 1;
NOINLINE uintptr_t validate(TxThread*);

struct RIV1 {
	static TM_FASTCALL bool begin(TxThread*);
	static TM_FASTCALL void commit(STM_COMMIT_SIG(,));
	static TM_FASTCALL void commit_ro(STM_COMMIT_SIG(,));
	static TM_FASTCALL void commit_rw(STM_COMMIT_SIG(,));
	static TM_FASTCALL void* read_ro(STM_READ_SIG(,,));
	static TM_FASTCALL void* read_rw(STM_READ_SIG(,,));
	static TM_FASTCALL void write_ro(STM_WRITE_SIG(,,,));
	static TM_FASTCALL void write_rw(STM_WRITE_SIG(,,,));
	static stm::scope_t* rollback(STM_ROLLBACK_SIG(,,,));
	static bool irrevoc(STM_IRREVOC_SIG(,));
	static void onSwitchTo();
};

/*
 * RIV1 Methods
 */

void* rcl_anon_mmap(size_t n) {
	void* res = mmap(0, n, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE,
			-1, 0);
	if (res == MAP_FAILED)
		fprintf(stderr, "Cannot allocate anon mmap");
	return res;
}

static void rcl_extract_topology(const char* cmd_nodes) {
	FILE *file;
	char text[1024], *p, *saveptr;
	int nb_nodes = 0;
	int nb_cores = 0;

	/* We use UNIX commands to find the nodes. */
	if (!(file = popen(cmd_nodes, "r")))
		fprintf(stderr, "Cannot open system file");

	/* For each node... */
	while (fgets(text, 1024, file)) {
		for (p = 0, p = strtok_r(text, " ", &saveptr); p;
				p = strtok_r(0, " ", &saveptr))
			nb_cores++;
		nb_nodes++;
	}

	printf("Number of nodes: %d\n", nb_nodes);
	printf("Number of cores: %d\n", nb_cores);

	num_cores = nb_cores;

	// close file
	if (pclose(file) < 0)
		fprintf(stderr, "Cannot close system file");

	// add first core to server cpu-set and all other cores to clients cpu-set
	// These cpu-sets will be used in threads binding
	for (int i = 0; i < nb_cores; i++) {

		// first core is dedicated for server and other cores are dedicated for clients.
		if (i == 0)
			CPU_SET(i, &main_server_cpuset);
		else
			CPU_SET(i, &client_cpuset);
	}
}

void* run_main_server(void* dummy) {

	// determine number of cores
	rcl_extract_topology(GET_NODES_CMD);

	// bind server thread to its core
	pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
			&main_server_cpuset);
	//printf("Main Server CPU is: %d\n", sched_getcpu());

	// Allocate "requests" Array
	size_t request_size =
			r_align(sizeof(struct request)*stm::MAX_THREADS, PAGE_SIZE);
	requests = (request*) rcl_anon_mmap(request_size);

	for (unsigned int i = 0; i < stm::MAX_THREADS; i++)
		requests[i].state = READY;

	// mark server as initialized so that clients can start
	server_initialized = true;

	volatile request* req;
	int i = 0, j = 0;

	// server loop
	while (clients_exist) {

		for (i = 0; i < threadcount.val; i++) {
			req = &(requests[i]);
			// pending transaction commit request
			if (req->state == PENDING) {
				// Validation is needed because client may be invalidated after changing its state to PENDING
				if (__builtin_expect(req->tx->alive == 2, false)) {
					req->state = ABORTED;
					continue;
				}

				uintptr_t tmp = timestamp.val;
				bcasptr(&timestamp.val, tmp, (tmp+1));
//				timestamp.val++;
//				CFENCE;

				// kill conflicting transactions
				for (j = 0; j < threadcount.val; j++)
					if ((i != j) && (threads[j]->alive == 1)
							&& (req->tx->wf->intersect(threads[j]->rf)))
						threads[j]->alive = 2;

				req->tx->writes.writeback(STM_WHEN_PROTECT_STACK(upper_stack_bound));
				req->tx->alive = 0;
				CFENCE;

				timestamp.val++;
				CFENCE;
				req->state = READY;
			}
		}
	}

	// deallocate request array
	munmap(requests, request_size);

	// clear flag when killing server
	server_initialized = false;

	printf("Server finished\n");

	return NULL;
}

/*
 * End of RIV1 server loop
 */

bool RIV1::irrevoc(STM_IRREVOC_SIG(tx,upper_stack_bound)) {

	fprintf(stderr, "Irrevoc is not yet implemented");

	return true;
}

void RIV1::onSwitchTo() {

	// We just need to be sure that the timestamp is not odd, or else we will
	// block.  For safety, increment the timestamp to make it even, in the event
	// that it is odd.

	if (timestamp.val & 1)
		++timestamp.val;

	// Create Server thread if not exist...
	if (!server_initialized) {
		void* args = (void*) (stm::MAX_THREADS);
		pthread_t tid;
		pthread_attr_t attr;
		pthread_attr_init(&attr);
		pthread_create(&tid, &attr, &run_main_server, args);

		// wait until server thread initializes its data-structures
		while (!server_initialized)
			;
	}
}

bool RIV1::begin(TxThread* tx) {

	// bind thread to clinet core-set
	if (!tx->bounded) {
		pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
				&client_cpuset);
		tx->bounded = true;
	}

	// Sample the sequence lock, if it is even decrement by 1
//	tx->start_time = timestamp.val & ~(1L);

	// notify the allocator
	tx->allocator.onTxBegin();
	tx->alive = 1;

	return false;
}

void RIV1::commit(STM_COMMIT_SIG(tx,upper_stack_bound)) {
	fprintf(stderr,
			"COMMIT IS NOT SUPPOSED TO BE CALLED. COMMIT_RO AND COMMIT_RW ARE ONLY USED");
}

void RIV1::commit_ro(STM_COMMIT_SIG(tx,)) {
	// if the transaction is invalid, abort
	if (__builtin_expect(tx->alive == 2, false))
		tx->tmabort(tx);

	// ok, all is good
	tx->alive = 0;
	tx->rf->clear();
	OnReadOnlyCommit(tx);

}

void RIV1::commit_rw(STM_COMMIT_SIG(tx,upper_stack_bound)) {

	// Validation. Saves time of servers on useless invalidated requests
	// if the transaction is invalid, abort
	if (__builtin_expect(tx->alive == 2, false))
		tx->tmabort(tx);

	// send request and then loop
	requests[tx->id - 1].tx = tx;
	requests[tx->id - 1].state = PENDING;
	while (requests[tx->id - 1].state == PENDING)
		;

	if (requests[tx->id - 1].state == ABORTED) {
		requests[tx->id - 1].state = READY;
		tx->tmabort(tx);
	} else {
		// clean up
		tx->writes.reset();
		tx->rf->clear();
		tx->wf->clear();
		OnReadWriteCommit(tx, read_ro, write_ro, commit_ro);
	}
}

void*
RIV1::read_ro(STM_READ_SIG(tx,addr,mask)) {
	// push address into read filter, ensure ordering w.r.t. the subsequent
	// read of data
	tx->rf->atomic_add(addr);

	// get a consistent snapshot of the value
	while (true) {
		uintptr_t x1 = timestamp.val;
		CFENCE;
		void* val = *addr;
		CFENCE;
		// if the ts was even and unchanged, then the read is valid
		bool ts_ok = !(x1 & 1) && (timestamp.val == x1);
		CFENCE;
		// if read valid, and we're not killed, return the value
		if ((tx->alive == 1) && ts_ok)
			return val;
		// abort if we're killed
		if (tx->alive == 2)
			tx->tmabort(tx);
	}
}

void*
RIV1::read_rw(STM_READ_SIG(tx,addr,mask)) {
	// check the log for a RAW hazard, we expect to miss
	WriteSetEntry log(STM_WRITE_SET_ENTRY(addr, NULL, mask));
	bool found = tx->writes.find(log);
	REDO_RAW_CHECK(found, log, mask);

	// reuse the ReadRO barrier, which is adequate here---reduces LOC
	void* val = read_ro(tx, addr STM_MASK(mask));
	REDO_RAW_CLEANUP(val, found, log, mask);
	return val;
}

void RIV1::write_ro(STM_WRITE_SIG(tx,addr,val,mask)) {
	// buffer the write, update the filter
	tx->writes.insert(WriteSetEntry(STM_WRITE_SET_ENTRY(addr, val, mask)));
	tx->wf->add(addr);
	OnFirstWrite(tx, read_rw, write_rw, commit_rw);
}

void RIV1::write_rw(STM_WRITE_SIG(tx,addr,val,mask)) {
	// just buffer the write
	tx->writes.insert(WriteSetEntry(STM_WRITE_SET_ENTRY(addr, val, mask)));
	tx->wf->add(addr);
}

stm::scope_t*
RIV1::rollback(STM_ROLLBACK_SIG(tx, upper_stack_bound, except, len)) {
	PreRollback(tx);

	// Perform writes to the exception object if there were any... taking the
	// branch overhead without concern because we're not worried about
	// rollback overheads.
	STM_ROLLBACK(tx->writes, upper_stack_bound, except, len);

	// clear filters and logs
	tx->rf->clear();
	if (tx->writes.size()) {
		tx->writes.reset();
		tx->wf->clear();
	}
	return PostRollback(tx, read_ro, write_ro, commit_ro);

}

} // (anonymous namespace)

namespace stm {
/**
 *  RIV1 initialization
 */
template<>
void initTM<RIV1>() {
	// set the name
	stms[RIV1].name = "RIV1";

	// set the pointers
	stms[RIV1].begin = ::RIV1::begin;
	stms[RIV1].commit = ::RIV1::commit_ro;
	stms[RIV1].read = ::RIV1::read_ro;
	stms[RIV1].write = ::RIV1::write_ro;
	stms[RIV1].rollback = ::RIV1::rollback;
	stms[RIV1].irrevoc = ::RIV1::irrevoc;
	stms[RIV1].switcher = ::RIV1::onSwitchTo;
	stms[RIV1].privatization_safe = true;
}
}
