/**
 *  Copyright (C) 2013
 *
 *  	 Implementation
 *
 * This is the implementation of the second version of remote invalidation.
 * In this version auxiliary threads are invalidating client threads
 * instead of the server thread.
 *
 */

#include "algs.hpp"
#include "RedoRAWUtils.hpp"
#include <sys/mman.h>

// Don't just import everything from stm. This helps us find bugs.
using stm::TxThread;
using stm::timestamp;
using stm::invalidators_timestamps;
using stm::client_cpuset;
using stm::main_server_cpuset;
using stm::invalidator_server_cpuset;
using stm::server_initialized;
using stm::num_cores;
using stm::commit_tx;
using stm::num_acks;
using stm::threadcount;
using stm::threads;
using stm::clients_exist;
using stm::WriteSetEntry;
using stm::ValueList;
using stm::ValueListEntry;
using stm::filter_t;
using stm::pad_word_t;
using stm::NUM_INVALIDATORS;
using stm::MAX_THREADS;

/*
 *  RIV3 definitions
 */

#define PAGE_SIZE	4096
#define CACHE_LINE_SIZE 64
#define r_align(n, r)        (((n) + (r) - 1) & -(r))
#define cache_align(n)       r_align(n , CACHE_LINE_SIZE)
#define pad_to_cache_line(n) (cache_align(n) - (n))

// Server-Client Request states
#define READY 		0
#define PENDING		1
#define ABORTED		2

#define GET_NODES_CMD																										\
		"NODES=/sys/devices/system/node;"																			\
		"CPUS=/sys/devices/system/cpu;"																				\
		"if [ -d $NODES ]; then"																							\
	  "  for F in `ls -d $NODES/node*`; do if [ $(cat $F/cpulist | grep '-') ]; then seq -s ' ' $(cat $F/cpulist | tr -s '-' ' '); else cat $F/cpulist | tr -s ',' ' '; fi; done;" \
		"else"																																\
		"  find $CPUS -maxdepth 1 -name \"cpu[0-9]*\" -exec basename {} \\; | sed -e 's/cpu\\(.*\\)/\\1/g' | tr -s '\n' ' ';" \
		"fi"

/*
 *  RIV3 structures
 */
struct request {
	volatile int state; /* the state of commit request */
	TxThread* tx; /* Tx context */
	char pad[pad_to_cache_line(sizeof(int) + sizeof(TxThread*))];
};

struct invalidate_data {
	volatile filter_t* commit_wf;
	volatile int commit_id;
	char pad[pad_to_cache_line(sizeof(int) + sizeof(filter_t*))];
};

/*
 * RIV3 Variables
 */

static struct request* requests = 0; /* requests array */
static struct invalidate_data* invalidate_array = 0; /* invalidate array */
static const uint32_t NUM_STEPS_AHEAD = 1;

namespace {

const uintptr_t VALIDATION_FAILED = 1;
NOINLINE uintptr_t validate(TxThread*);

struct RIV3 {
	static TM_FASTCALL bool begin(TxThread*);
	static TM_FASTCALL void commit(STM_COMMIT_SIG(,));
	static TM_FASTCALL void commit_ro(STM_COMMIT_SIG(,));
	static TM_FASTCALL void commit_rw(STM_COMMIT_SIG(,));
	static TM_FASTCALL void* read_ro(STM_READ_SIG(,,));
	static TM_FASTCALL void* read_rw(STM_READ_SIG(,,));
	static TM_FASTCALL void write_ro(STM_WRITE_SIG(,,,));
	static TM_FASTCALL void write_rw(STM_WRITE_SIG(,,,));
	static stm::scope_t* rollback(STM_ROLLBACK_SIG(,,,));
	static bool irrevoc(STM_IRREVOC_SIG(,));
	static void onSwitchTo();
};

/*
 * RIV3 Methods
 */

void* rcl_anon_mmap(size_t n) {
	void* res = mmap(0, n, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE,
			-1, 0);
	if (res == MAP_FAILED)
		fprintf(stderr, "Cannot allocate anon mmap");
	return res;
}

static void rcl_extract_topology(const char* cmd_nodes) {
	FILE *file;
	char text[1024], *p, *saveptr;
	int nb_nodes = 0;
	int nb_cores = 0;

	/* We use UNIX commands to find the nodes. */
	if (!(file = popen(cmd_nodes, "r")))
		fprintf(stderr, "Cannot open system file");

	/* For each node... */
	while (fgets(text, 1024, file)) {
		for (p = 0, p = strtok_r(text, " ", &saveptr); p;
				p = strtok_r(0, " ", &saveptr))
			nb_cores++;
		nb_nodes++;
	}

	printf("Number of nodes: %d\n", nb_nodes);
	printf("Number of cores: %d\n", nb_cores);

	num_cores = nb_cores;

	// close file
	if (pclose(file) < 0)
		fprintf(stderr, "Cannot close system file");

	// add first core to server cpu-set and all other cores to clients cpu-set
	// These cpu-sets will be used in threads binding
	for (int i = 0; i < nb_cores; i++) {

		// first core is dedicated for server and other cores are dedicated for clients.
		if (i == 0)
			CPU_SET(i, &main_server_cpuset);
		else if (i <= NUM_INVALIDATORS)
			CPU_SET(i, &invalidator_server_cpuset);
		else
			CPU_SET(i, &client_cpuset);
	}
}

void* run_main_server(void* dummy) {

	// determine number of cores
	rcl_extract_topology(GET_NODES_CMD);

	// bind server thread to its core
	pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
			&main_server_cpuset);
	//printf("Main Server CPU is: %d\n", sched_getcpu());

//	commit_wf = (filter_t*) FILTER_ALLOC(sizeof(filter_t));

	// Allocate "requests" Array
	size_t request_size =
			r_align(sizeof(struct request)*MAX_THREADS, PAGE_SIZE);
	requests = (request*) rcl_anon_mmap(request_size);

	// Allocate "invalidate" Array
	size_t invalidate_size =
			r_align(sizeof(struct invalidate_data)*MAX_THREADS, PAGE_SIZE);
	invalidate_array = (invalidate_data*) rcl_anon_mmap(invalidate_size);

//	// Allocate "application-Auxiliary" Array
//	size_t app_aux_size = r_align(sizeof(int)*MAX_THREADS, PAGE_SIZE);
//	app_aux_array = (int*) rcl_anon_mmap(app_aux_size);

	// initialize variables
	for (unsigned int i = 0; i < MAX_THREADS; i++) {
		requests[i].state = READY;
		invalidate_array[i].commit_wf = (filter_t*) FILTER_ALLOC(sizeof(filter_t));
	}

	CFENCE;
	// mark server as initialized so that clients can start
	server_initialized = true;

	volatile request* req;
	int i = 0, j = 0, index = 0;
	// server loop
	while (clients_exist) {

		for (i = 0; i < threadcount.val; i++) {
			req = &(requests[i]);
			// pending transaction commit request
			if (req->state == PENDING && invalidators_timestamps[req->tx->invalidator_id].val >= timestamp.val) {

				for (j = 0; j < NUM_INVALIDATORS; j++) {
					while (invalidators_timestamps[j].val + NUM_STEPS_AHEAD < timestamp.val)
						;
				}
				CFENCE;

				// Validation is needed because client may be invalidated after changing its state to PENDING
				if (__builtin_expect(req->tx->alive == 2, false)) {
					req->state = ABORTED;
					continue;
				}

				index = timestamp.val%MAX_THREADS;
				invalidate_array[index].commit_wf->fastcopy(req->tx->wf);
				invalidate_array[index].commit_id = req->tx->id;
				CFENCE;

				timestamp.val++;
				CFENCE;

				req->tx->writes.writeback(STM_WHEN_PROTECT_STACK(upper_stack_bound));
				req->tx->alive = 0;

				CFENCE;
				timestamp.val++;

				req->state = READY;
			}
		}
	}

	// deallocate arrays
	munmap(requests, request_size);
	munmap(invalidate_array, invalidate_size);

	// clear flag when killing server
	server_initialized = false;

	printf("Server finished\n");

	return NULL;
}

/*
 * End of RIV3 server loop
 */

void* run_invalidator(void* arg) {

	// bind server thread to its core
	pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
			&invalidator_server_cpuset);

	volatile int invalidator_id = *((int*) arg);
	uint32_t i, index;

	while (clients_exist) {
		if (timestamp.val > invalidators_timestamps[invalidator_id].val) {
			index = invalidators_timestamps[invalidator_id].val%MAX_THREADS;
			// wait for all live threads to finish invalidation
			for (i = invalidator_id; i < threadcount.val; i +=
					NUM_INVALIDATORS) {
				if ((threads[i]->id != invalidate_array[index].commit_id)
						&& (threads[i]->alive == 1)
						&& (invalidate_array[index].commit_wf->intersect(threads[i]->rf)))
					threads[i]->alive = 2;
				threads[i]->start_time = invalidators_timestamps[invalidator_id].val + 2;
			}
			CFENCE;

			invalidators_timestamps[invalidator_id].val+=2;
			CFENCE;

		}
	}

	return NULL;
}

bool RIV3::irrevoc(STM_IRREVOC_SIG(tx,upper_stack_bound)) {

	fprintf(stderr, "Irrevoc is not yet implemented");

	return true;
}

void RIV3::onSwitchTo() {

	// We just need to be sure that the timestamp is not odd, or else we will
	// block.  For safety, increment the timestamp to make it even, in the event
	// that it is odd.

	if (timestamp.val & 1)
		++timestamp.val;
	invalidators_timestamps = (pad_word_t*) rcl_anon_mmap(
			sizeof(pad_word_t) * NUM_INVALIDATORS);
	for (int j = 0; j < NUM_INVALIDATORS; j++)
		invalidators_timestamps[j].val = timestamp.val;
	CFENCE;

	// Create Server thread if not exist...
	if (!server_initialized) {

		void* args = (void*) (MAX_THREADS);
		pthread_t tid;
		pthread_attr_t attr;
		pthread_attr_init(&attr);
		pthread_create(&tid, &attr, &run_main_server, args);

		// wait until server thread initializes its data-structures
		while (!server_initialized)
			;

		for (int j = 0; j < NUM_INVALIDATORS; j++) {
			int* invalidator_id = (int*) malloc(sizeof(int));
			*invalidator_id = j;
			void* invalidator_args = (void*) invalidator_id;
			pthread_t invalidator_tid;
			pthread_attr_t invalidator_attr;
			pthread_attr_init(&invalidator_attr);
			pthread_create(&invalidator_tid, &invalidator_attr,
					&run_invalidator, invalidator_args);
		}
	}
}

bool RIV3::begin(TxThread* tx) {

	// bind thread to clinet core-set
	if (!tx->bounded) {
		pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
				&client_cpuset);
		tx->bounded = true;
	}

	// notify the allocator
	tx->allocator.onTxBegin();
	tx->start_time = invalidators_timestamps[tx->invalidator_id].val;
	tx->alive = 1;
	return false;
}

void RIV3::commit(STM_COMMIT_SIG(tx,upper_stack_bound)) {
	fprintf(stderr,
			"COMMIT IS NOT SUPPOSED TO BE CALLED. COMMIT_RO AND COMMIT_RW ARE ONLY USED");
}

void RIV3::commit_ro(STM_COMMIT_SIG(tx,)) {
	// if the transaction is invalid, abort
	if (__builtin_expect(tx->alive == 2, false))
		tx->tmabort(tx);

	// ok, all is good
	tx->alive = 0;
	tx->rf->clear();
	OnReadOnlyCommit(tx);

}

void RIV3::commit_rw(STM_COMMIT_SIG(tx,upper_stack_bound)) {

	//loop until finishing all the pending commits not handled by auxiliary thread yet
//	while (timestamp.val != invalidators_timestamps[tx->invalidator_id].val)
//		;

	// Validation. Saves time of servers on useless invalidated requests
	// if the transaction is invalid, abort
	if (__builtin_expect(tx->alive == 2, false))
		tx->tmabort(tx);

	// send request and then loop
	requests[tx->id - 1].tx = tx;
	requests[tx->id - 1].state = PENDING;
	while (requests[tx->id - 1].state == PENDING)
		;

	if (requests[tx->id - 1].state == ABORTED) {
		requests[tx->id - 1].state = READY;
		tx->tmabort(tx);
	} else {
		// clean up
		tx->writes.reset();
		tx->rf->clear();
		tx->wf->clear();
		OnReadWriteCommit(tx, read_ro, write_ro, commit_ro);
	}
}

void*
RIV3::read_ro(STM_READ_SIG(tx,addr,mask)) {
	// push address into read filter, ensure ordering w.r.t. the subsequent
	// read of data
	tx->rf->atomic_add(addr);

	// get a consistent snapshot of the value
	while (true) {

		//loop until finishing all the pending commits not handled by auxiliary thread yet
		uintptr_t x1 = timestamp.val;
		CFENCE;
		void* val = *addr;
		CFENCE;
		// if the ts was even and unchanged, then the read is valid
		// NOTE: x1 has to be even because start_time is always even
		bool ts_ok = (timestamp.val == x1) && !(x1 & 1)
				&& (invalidators_timestamps[tx->invalidator_id].val == x1)
				//&& (tx->start_time == x1)
						;
		CFENCE;
		// if read valid, and we're not killed, return the value
		if ((tx->alive == 1) && ts_ok)
			return val;
		// abort if we're killed
		if (tx->alive == 2)
			tx->tmabort(tx);
	}
}

void*
RIV3::read_rw(STM_READ_SIG(tx,addr,mask)) {
	// check the log for a RAW hazard, we expect to miss
	WriteSetEntry log(STM_WRITE_SET_ENTRY(addr, NULL, mask));
	bool found = tx->writes.find(log);
	REDO_RAW_CHECK(found, log, mask);

	// reuse the ReadRO barrier, which is adequate here---reduces LOC
	void* val = read_ro(tx, addr STM_MASK(mask));
	REDO_RAW_CLEANUP(val, found, log, mask);
	return val;
}

void RIV3::write_ro(STM_WRITE_SIG(tx,addr,val,mask)) {
	// buffer the write, update the filter
	tx->writes.insert(WriteSetEntry(STM_WRITE_SET_ENTRY(addr, val, mask)));
	tx->wf->add(addr);
	OnFirstWrite(tx, read_rw, write_rw, commit_rw);
}

void RIV3::write_rw(STM_WRITE_SIG(tx,addr,val,mask)) {
	// just buffer the write
	tx->writes.insert(WriteSetEntry(STM_WRITE_SET_ENTRY(addr, val, mask)));
	tx->wf->add(addr);
}

stm::scope_t*
RIV3::rollback(STM_ROLLBACK_SIG(tx, upper_stack_bound, except, len)) {
	PreRollback(tx);

	// Perform writes to the exception object if there were any... taking the
	// branch overhead without concern because we're not worried about
	// rollback overheads.
	STM_ROLLBACK(tx->writes, upper_stack_bound, except, len);

	// clear filters and logs
	tx->rf->clear();
	if (tx->writes.size()) {
		tx->writes.reset();
		tx->wf->clear();
	}
	return PostRollback(tx, read_ro, write_ro, commit_ro);

}

} // (anonymous namespace)

namespace stm {
/**
 *  RIV3 initialization
 */
template<>
void initTM<RIV3>() {
	// set the name
	stms[RIV3].name = "RIV3";

	// set the pointers
	stms[RIV3].begin = ::RIV3::begin;
	stms[RIV3].commit = ::RIV3::commit_ro;
	stms[RIV3].read = ::RIV3::read_ro;
	stms[RIV3].write = ::RIV3::write_ro;
	stms[RIV3].rollback = ::RIV3::rollback;
	stms[RIV3].irrevoc = ::RIV3::irrevoc;
	stms[RIV3].switcher = ::RIV3::onSwitchTo;
	stms[RIV3].privatization_safe = true;
}
}
